#!/bin/bash
#
# Copyright (C) 2009 Väinö Helminen <vah78@yahoo.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#

function update()
{
  local path=`readlink -e .`
  : > .md5

  if ! [ -r .md5 ] ; then 
    echo "!!! Will not proceed to this directory: $path"
    return
  fi

  # build lists of all files and directories
  local dirlist=`mktemp`
  local filelist=`mktemp`
  for file in * ; do 
    # ignore symbolic links
    [ -h "$file" ] && continue

    if [ -f "$file" ] ; then 
      echo "$file" >> $filelist
    elif [ -d "$file" ] ; then
      echo "$file" >> $dirlist
    fi
  done

  # go through the sorted list of filenames
  sort  $filelist | while read file ; do 
    if [ -r "$file" ] ; then
      md5=`md5sum -- "$file"|cut -d ' ' -f1`
      file=`echo $file` # remove newlines
      echo "$md5  $file" >> .md5
      echo "$md5  $path/$file"
    else
      echo "!!! Unreadble file: $path/$file"
    fi
  done
  rm -f $filelist
  unset filelist

  # recurse into subdirectories (sorted list, no duplicates)
  sort $dirlist | while read file ; do   
    local dir=$file
    if cd "$dir" 2>/dev/null ; then
      update
      cd "$path"
    else
      echo "!!! Inaccessible directory: $path/$dir"
    fi
  done
  rm -f $dirlist
}


function check()
{
  local path=`readlink -e .`

  # build lists of all files and directories
  local dirlist=`mktemp`
  local filelist=`mktemp`
  for file in * ; do 
    # ignore symbolic links
    [ -h "$file" ] && continue

    if [ -f "$file" ] ; then 
      echo "$file" >> $filelist
    elif [ -d "$file" ] ; then
      echo "$file" >> $dirlist
    fi
  done

  if [ -r .md5 ] ; then
    # add files with saved hashes to the hash
    sed 's/^[0-9a-f]\+\s\+//' < .md5 >> $filelist

    # go through the sorted list of filenames with duplicates removed
    sort -u $filelist | while read file ; do 
      if [ -f "$file" ] ; then
        if [ -r "$file" ] ; then
          md5=`md5sum -- "$file" | cut -d' ' -f1`
          line=`fgrep -x "$md5  $file" < .md5`
          if [ -z "$line" ] ; then 
            if sed 's/^[0-9a-f]\+\s\+//' < .md5 | fgrep -xq "$file" ; then
	      echo "***$md5  $path/$file" # file's checksum has changed
            else
	      echo "+++$md5  $path/$file" # new file
            fi
          else
            echo "$md5  $path/$file" # OK, old and unchanged file
          fi
        else
          echo "!!! Unreadble file: $path/$file"
        fi
      else
        fgrep "  $file" < .md5 | while read line ; do 
          md5=`echo "$line" | cut -d' ' -f1`
          if fgrep -xq "$md5  $file" < .md5 ; then
            echo "---$md5  $path/$file" # file has been removed
            break;
          fi
        done
      fi
    done
  else
    echo "!!! No checksum file (.md5): $path"
  fi
  rm -f $filelist
  unset filelist

  # recurse into subdirectories (sorted list, no duplicates)
  sort $dirlist | while read file ; do   
    local dir=$file
    if cd "$dir" 2>/dev/null ; then
      check
      cd "$path"
    else
      echo "!!! Inaccessible directory: $path/$dir"
    fi
  done
  rm -f $dirlist
}


function dump()
{
  local path=`readlink -e .`

  # dump sorted list of stored hashes of files that should be in this directory
  if [ -r .md5 ] ; then 
    sort -k 2 < .md5 | awk -v path=" $path/" '{$2=path$2 ;print }'
  else
    echo "!!! No checksum file (.md5): $path"
  fi

  # recurse into subdirectories (in alphabetical order)
  for file in * ; do 
    [ ! -h "$file" -a  -d "$file" ] && echo "$file"
  done | sort $dirlist | while read file ; do 
    [ -h "$file" ] && continue
    if [ -d "$file" ] ; then
      local dir="$file"
      if cd "$dir" 2>/dev/null; then
        dump
        cd "$path"
      else
        echo "!!! Inaccessible directory: $path/$dir"
      fi
    fi
  done
}

function list()
{
  local path=`readlink -e .`

  # build lists of all files and directories
  local dirlist=`mktemp`
  local filelist=`mktemp`
  for file in * ; do 
    # ignore symbolic links
    [ -h "$file" ] && continue

    if [ -f "$file" ] ; then 
      echo "$file" >> $filelist
    elif [ -d "$file" ] ; then
      echo "$file" >> $dirlist
    fi
  done

  # go through the sorted list of filenames
  sort  $filelist | while read file ; do 
    if [ -r "$file" ] ; then
      md5=`md5sum -- "$file"|cut -d ' ' -f1`
      file=`echo $file` # remove newlines
      echo "$md5  $path/$file"
    else
      echo "!!! Unreadble file: $path/$file"
    fi
  done
  rm -f $filelist
  unset filelist

  # recurse into subdirectories (sorted list, no duplicates)
  sort $dirlist | while read file ; do   
    local dir=$file
    if cd "$dir" 2>/dev/null ; then
      list
      cd "$path"
    else
      echo "!!! Inaccessible directory: $path/$dir"
    fi
  done
  rm -f $dirlist
}

function help()
{
  cmd=`basename $0`
  cat <<EOF
Usage: $cmd -u|-c|-d|-l [directory]

  -u/--update  - calculate checksums of all files and the them to
                 a per directorory file named ".md5"

  -c/--check   - verify files against the saved checksums

  -d/--dump    - generate a list of all saved checksums

  -l/--list    - generate a list of current checksums

  -h/--help/gibberish - show this help message

Notes:
  $cmd always works recursively.

  The saved .md5-files are compatible with the md5sum command and can be
  generated also with "md5sum -- * > .md5" and checked with
  "md5sum -c .md5" for each directory.
EOF
}


if ! md5sum "$0" &>/dev/null ; then
  echo 'Sorry, cannot find the "md5sum" command.' >&2
  exit 1;
fi

if [ -n "$2" ] ; then 
  cd "$2" || exit 1
fi

case $1 in
-u | --update)
  update;;
-c | --check | --verify)
  check;;
-d | --dump)
  dump;;
-l | --list)
  list;;
*)
  help;;
esac
